import numpy as np
from scipy.interpolate import interp1d


def confidence(F, alpha=.05):
    r"""
    Constructs a Dvoretzky-Kiefer-Wolfowitz confidence band for the eCDF.

    Parameters
    ----------
    F : array_like
        The empirical distributions
    alpha : float
        Set alpha for a (1 - alpha) % confidence band.

    Notes
    -----
    Based on the DKW inequality.

    .. math:: P \left( \sup_x \left| F(x) - \hat(F)_n(X) \right| > \epsilon \right) \leq 2e^{-2n\epsilon^2}

    References
    ----------
    Wasserman, L. 2006. `All of Nonparametric Statistics`. Springer.

    """
    nobs = len(F)
    epsilon = np.sqrt(np.log(2./alpha) / (2 * nobs))
    lower = np.clip(F - epsilon, 0, 1)
    upper = np.clip(F + epsilon, 0, 1)

    return lower, upper


def get_cdf(x):
    _x = np.array(x, copy=True)
    _x.sort()
    _y = np.linspace(1./len(_x), 1, len(_x))

    _x = np.r_[-np.inf, _x]
    _y = np.r_[0, _y]

    def cdf(z):
        tind = np.searchsorted(_x, z, 'right') - 1
        return _y[tind]

    return cdf


def kron(a, b):
    return np.kron(a, b)


def orthogonalize(Y, mu):
    L = np.array([[1]])
    R = np.array([[1]])
    for i in range(0, mu):
        R = orthogonalize_left(Y, i)
    for i in range(len(Y)-1, mu, -1):
        L = orthogonalize_right(Y, i)
    return R, L


def orthogonalize_left(Y, mu):
    assert 0 <= mu < len(Y)-1
    G = unfolding_left(Y[mu])
    Q, R = np.linalg.qr(G, mode='reduced')
    Y[mu] = reshape(Q, Y[mu].shape[:-1] + (Q.shape[1], ))
    G = unfolding_right(Y[mu+1])
    Y[mu+1] = reshape(np.dot(R, G), (R.shape[0], ) + Y[mu+1].shape[1:])
    return R


def orthogonalize_right(Y, mu):
    assert 1 <= mu < len(Y)
    G = unfolding_right(Y[mu])
    L, Q = scipy.linalg.rq(G, mode='economic', check_finite=False)
    Y[mu] = reshape(Q, (Q.shape[0], ) + Y[mu].shape[1:])
    G = unfolding_left(Y[mu-1])
    Y[mu-1] = reshape(np.dot(G, L), Y[mu-1].shape[:-1] + (L.shape[1], ))
    return L


def reshape(a, sz):
    return np.reshape(a, sz, order = 'F')


def svd_truncated(M, delta=None, eps=None, rmax=None, left_ortho=True):
    if delta is not None and eps is not None:
        raise ValueError('Provide either `delta` or `eps`')
    if delta is None and eps is not None:
        delta = eps*np.linalg.norm(M)
    if delta is None and eps is None:
        delta = 0
    if rmax is None:
        rmax = np.iinfo(np.int32).max
    assert rmax >= 1

    if M.shape[0] <= M.shape[1]:
        cov = M.dot(M.T)
        singular_vectors = 'left'
    else:
        cov = M.T.dot(M)
        singular_vectors = 'right'

    if np.linalg.norm(cov) < 1e-14:
        return np.zeros([M.shape[0], 1]), np.zeros([1, M.shape[1]])

    w, v = np.linalg.eigh(cov)
    w[w < 0] = 0
    w = np.sqrt(w)
    svd = [v, w]
    # Sort eigenvalues and eigenvectors in decreasing importance
    idx = np.argsort(svd[1])[::-1]
    svd[0] = svd[0][:, idx]
    svd[1] = svd[1][idx]

    S = svd[1]**2
    where = np.where(np.cumsum(S[::-1]) <= delta**2)[0]
    if len(where) == 0:
        rank = max(1, int(np.min([rmax, len(S)])))
    else:
        rank = max(1, int(np.min([rmax, len(S) - 1 - where[-1]])))
    left = svd[0]
    left = left[:, :rank]

    if singular_vectors == 'left':
        if left_ortho:
            M2 = left.T.dot(M)
        else:
            M2 = ((1. / svd[1][:rank])[:, np.newaxis]*left.T).dot(M)
            left = left*svd[1][:rank]
    else:
        if left_ortho:
            M2 = M.dot(left * (1. / svd[1][:rank])[np.newaxis, :])
            left, M2 = M2, left.dot(np.diag(svd[1][:rank])).T
        else:
            M2 = M.dot(left)
            left, M2 = M2, left.T

    return left, M2


def unfolding_left(G):
    return reshape(G, [-1, G.shape[2]])


def unfolding_right(G):
    return reshape(G, [G.shape[0], -1])



import numpy as np


from .maxvol import maxvol
from .maxvol import rect_maxvol
from .tensor import getter
from .tensor import erank
from .utils import kron
from .utils import reshape


class Node:
    def __init__(self, core=None, edge1=None, edge2=None):
        self.core = core
        self.edges = [edge1, edge2]


class Edge:
    def __init__(self, node1=None, node2=None):
        self.nodes = [node1, node2]
        self.Ru = []
        self.Rv = []


class Tree:
    def __init__(self, d, Y, fun=None):
        self.d = d
        self.fun = fun
        self.edges = [Edge() for i in range(d + 1)]
        self.nodes = [Node(Y[i].copy(), self.edges[i], self.edges[i+1]) for i in range(d)]
        for i in range(d - 1):
            self.edges[i].nodes[1] = self.nodes[i]
            self.edges[i + 1].nodes[0] = self.nodes[i]


def cross_tmp(func, Y0, nswp=10, kickrank=2, rf=2.0, eps=None, val_size=10000, with_wrap=False, verbose=False):

    def func_wrapper(J):
        if with_wrap:
            Y = [c.nodes[i].core.copy() for i in range(c.d)]
            return func(J, Y)
        else:
            return func(J)

    d = len(Y0)
    c = Tree(d, Y0, func_wrapper)
    c.edges[0].Ru = np.ones((1, 1))
    c.edges[d].Rv = np.ones((1, 1))
    c.edges[0].ind_left = np.empty((1, 0), dtype=np.int32)
    c.edges[d].ind_right = np.empty((1, 0), dtype=np.int32)

    # setup_indices :
    for i in range(d-1):
        _left_qr_maxvol(c.nodes[i])
    for i in range(d-1, 0, -1):
        _right_qr_maxvol(c.nodes[i])

    #xold = [G.copy() for G in Y0]
    if verbose or eps is not None:
        # Create a validation set
        Xs_val = np.array([np.random.choice(I.shape[1], int(val_size)) for I in Y0], dtype=int).T
        ys_val = func(Xs_val)
        norm_ys_val = np.linalg.norm(ys_val)

    for s in range(nswp):
        for i in range(d):
            c.nodes[i].edges[0].Rv = np.ones((1, 1))
            c.nodes[i].edges[1].Rv = np.ones((1, 1))
            c.nodes[i].edges[0].Ru = np.ones((1, 1))
            c.nodes[i].edges[1].Ru = np.ones((1, 1))
        for i in range(d):
            _apply_Ru(c.nodes[i])
            c.fun_eval += _update_core_left(c.nodes[i], c.fun, kickrank, rf)

        c.nodes[d-1].core = np.tensordot(
            c.nodes[d-1].core, c.nodes[d-1].edges[1].Ru, 1)

        for i in range(c.d-1, -1, -1):
            _apply_Rv(c.nodes[i])
            c.fun_eval += _update_core_right(c.nodes[i], c.fun, kickrank, rf)

        c.nodes[0].core = np.tensordot(
            c.nodes[0].edges[0].Rv, c.nodes[0].core, 1)

        x1 = [c.nodes[i].core for i in range(c.d)]


        if verbose or eps is not None:
            get = getter(x1)
            y = np.array([get(x) for x in Xs_val])
            er = np.linalg.norm(ys_val - y)
            er_rel = er / norm_ys_val
        if verbose:
            # nrm = norm(x1)
            # er = (tt.tensor.from_list(x1) - tt.tensor.from_list(xold)).norm()
            # er_rel = er / nrm
            # print('swp: %d/%d er_rel = %3.1e er_abs = %3.1e erank = %3.1f fun_eval: %d' % (s, nswp-1, er_rel, er, erank(x1), c.fun_eval))
            print(f'swp: {s}/{nswp-1}, er_rel = {er_rel:3.1e}, er_abs = {er:3.1e}, erank = {erank(x1):3.1f}, fun_eval: {c.fun_eval}')
        if eps is not None and er_rel < eps:
            break
        # xold = [G.copy() for G in x1]

    return x1


def cross(func, Y0, nswp=10, kickrank=2, rf=2, wrap=False):

    def func_wrapper(J):
        if wrap:
            return func(J, [c.nodes[i].core.copy() for i in range(c.d)])
        else:
            return func(J)

    d = len(Y0)
    c = Tree(d, Y0, func_wrapper)
    _init(c)
    for _ in range(nswp):
        _prep(c)

        for i in range(d):
            _apply_Ru(c.nodes[i])
            _update_core_left(c.nodes[i], c.fun, kickrank, rf)

        c.nodes[d-1].core = np.tensordot(c.nodes[d-1].core, c.nodes[d-1].edges[1].Ru, 1)

        for i in range(c.d-1, -1, -1):
            _apply_Rv(c.nodes[i])
            _update_core_right(c.nodes[i], c.fun, kickrank, rf)

        c.nodes[0].core = np.tensordot(c.nodes[0].edges[0].Rv,c.nodes[0].core,1)

    return [c.nodes[i].core.copy() for i in range(c.d)]


def _init(c):
    c.edges[0].Ru = np.ones((1, 1))
    c.edges[c.d].Rv = np.ones((1, 1))
    c.edges[0].ind_left = np.empty((1, 0), dtype=np.int32)
    c.edges[c.d].ind_right = np.empty((1, 0), dtype=np.int32)
    for i in range(c.d - 1):
        _left_qr_maxvol(c.nodes[i])
    for i in range(c.d - 1, 0, -1):
        _right_qr_maxvol(c.nodes[i])


def _prep(c):
    for i in range(c.d):
        c.nodes[i].edges[0].Rv = np.ones((1, 1))
        c.nodes[i].edges[1].Rv = np.ones((1, 1))
        c.nodes[i].edges[0].Ru = np.ones((1, 1))
        c.nodes[i].edges[1].Ru = np.ones((1, 1))


def _index_merge(i1, i2, i3):
    if i1 is not []:
        r1 = i1.shape[0]
    else:
        r1 = 1
    r2 = i2.shape[0]
    if i1 is not []:
        r3 = i3.shape[0]
    else:
        r3 = 1
    w1 = kron(np.ones((r3 * r2, 1), dtype=np.int32), i1)
    try:
        w2 = kron(i3, np.ones((r1 * r2, 1), dtype=np.int32))
    except:
        w2 = np.empty((w1.shape[0], 0))
    w3 = kron(kron(np.ones((r3, 1), dtype=np.int32), i2), np.ones((r1, 1), dtype=np.int32))
    return np.hstack((w1, w3, w2))


def _left_qr_maxvol(nd):
    cr = nd.core.copy()
    r1, n1, r2 = cr.shape
    cr = np.tensordot(nd.edges[0].Ru, cr, 1)
    r1 = cr.shape[0]
    cr = reshape(cr, (r1 * n1, r2))
    q, Ru = np.linalg.qr(cr)
    ind, c = maxvol(q)
    Ru = np.dot(q[ind, :], Ru)
    q = c.copy()
    nd.core = reshape(q, (r1, n1, r2)).copy()
    nd.edges[1].Ru = Ru.copy()
    nd.maxvol_left = np.unravel_index(ind, (r1, n1), order='F')
    i_left = nd.edges[0].ind_left
    w1 = kron(np.ones((n1, 1), dtype=np.int32), i_left)
    w2 = kron(reshape(np.arange(n1, dtype=np.int32),(-1, 1)), np.ones((r1, 1), dtype=np.int32))
    i_next = np.hstack((w1, w2))
    i_next = reshape(i_next, (r1 * n1, -1))
    i_next = i_next[ind, :]

    nd.edges[1].ind_left = i_next.copy()
    nd.edges[1].ind_left_add = i_next.copy()


def _right_qr_maxvol(nd):
    cr = nd.core.copy()
    r1, n, r2 = cr.shape
    Rv = nd.edges[1].Rv.copy()
    cr = np.tensordot(cr, Rv, 1)
    r2 = cr.shape[2]
    cr = reshape(cr, (r1, -1))
    cr = cr.T
    q, Rv = np.linalg.qr(cr)
    ind, c = maxvol(q)
    Rv = np.dot(q[ind, :], Rv)
    q = c.copy()
    nd.edges[0].Rv = Rv.T.copy()
    q = reshape(q.T, (-1, n, r2))
    nd.core = q.copy()
    nd.maxvol_right = np.unravel_index(ind, (n, r2), order='F')
    i_right = nd.edges[1].ind_right
    w1 = kron(np.ones((r2, 1), dtype=np.int32), reshape(np.arange(n, dtype=np.int32),(-1, 1)))
    try:
        w2 = kron(i_right, np.ones((n, 1), dtype=np.int32))
        i_next = np.hstack((w1, w2))
    except:
        i_next = w1

    i_next = reshape(i_next, (n * r2, -1))
    i_next = i_next[ind, :]
    nd.edges[0].ind_right = i_next.copy()
    nd.edges[0].ind_right_add = i_next.copy()


def _apply_Ru(nd):
    cr = nd.core.copy()
    Ru = nd.edges[0].Ru.copy()
    cr = np.tensordot(Ru, cr, 1)
    nd.core = cr
    nd.edges[0].Ru = []


def _apply_Rv(nd):
    cr = nd.core.copy()
    Rv = nd.edges[1].Rv.copy()
    cr = np.tensordot(cr, Rv, 1)
    nd.core = cr
    nd.edges[1].Rv = []


def _update_core_left(nd, fun, kickrank, rf, tau=1.1):
    cr = nd.core.copy()
    r1, n, r2 = cr.shape
    i_left = nd.edges[0].ind_left
    i_right = nd.edges[1].ind_right
    p = np.arange(n, dtype=np.int32); p = reshape(p, (-1, 1))
    J = _index_merge(i_left, p, i_right)
    cr = fun(J)
    cr = reshape(cr, (-1, r2))

    q, s1, v1 = np.linalg.svd(cr, full_matrices=False)
    R = np.diag(s1).dot(v1)

    N, r = q.shape
    N_min = min(N, r + kickrank)
    N_max = min(N, r + kickrank + rf)
    if N <= r:
        ind_new = np.arange(q.shape[0], dtype=np.int32)
        C = np.eye(q.shape[0], dtype=q.dtype)
    else:
        ind_new, C = rect_maxvol(q, tau, N_min, N_max)
    nd.core = C
    nd.core = reshape(nd.core, (r1, n, -1))
    Ru = q[ind_new, :].dot(R)
    i_left = nd.edges[0].ind_left
    w1 = kron(np.ones((n, 1), dtype=np.int32), i_left)
    w2 = kron(reshape(np.arange(n, dtype=np.int32), (-1, 1)), np.ones((r1, 1), dtype=np.int32))

    i_next = np.hstack((w1, w2))

    i_next = reshape(i_next, (r1 * n, -1))

    nd.edges[1].ind_left = i_next[ind_new, :].copy()
    try:
        nd.edges[1].Ru = np.dot(Ru, nd.edges[1].Ru)
    except:
        nd.edges[1].Ru = Ru.copy()


def _update_core_right(nd, fun, kickrank, rf, tau=1.1):
    cr = nd.core
    r1, n, r2 = cr.shape

    i_left = nd.edges[0].ind_left
    i_right = nd.edges[1].ind_right

    p = np.arange(n, dtype=np.int32); p = reshape(p, (-1, 1))

    J = _index_merge(i_left, p, i_right)

    cr = fun(J)
    cr = reshape(cr, (r1, -1))
    cr = cr.T
    q, s1, v1 = np.linalg.svd(cr, full_matrices=False)
    R = np.diag(s1).dot(v1)

    N, r = q.shape
    N_min = min(N, r + kickrank)
    N_max = min(N, r + kickrank + rf)
    if N <= r:
        ind_new = np.arange(q.shape[0], dtype=np.int32)
        C = np.eye(q.shape[0], dtype=q.dtype)
    else:
        ind_new, C = rect_maxvol(q, tau, N_min, N_max)
    nd.core = C
    nd.core = reshape(nd.core.T, (-1, n, r2))
    Rv = q[ind_new, :].dot(R)

    try:
        nd.edges[0].Rv = np.dot(nd.edges[0].Rv, Rv.T)
    except:
        nd.edges[0].Rv = Rv.T.copy()

    i_right = nd.edges[1].ind_right
    w1 = kron(np.ones((r2, 1), dtype=np.int32), reshape(np.arange(n, dtype=np.int32),(-1, 1)))
    try:
        w2 = kron(i_right, np.ones((n, 1), dtype=np.int32))
        i_next = np.hstack((w1, w2))
    except:
        i_next = w1

    i_next = reshape(i_next, (n * r2, -1))
    nd.edges[0].ind_right = i_next[ind_new, :].copy()



from .cross import cross


from .maxvol import maxvol
from .maxvol import rect_maxvol


from .tensor import erank
from .tensor import get
from .tensor import getter
from .tensor import mean
from .tensor import mul
from .tensor import norm
from .tensor import rand
from .tensor import recap
from .tensor import truncate


from .utils import confidence
from .utils import get_cdf



import numpy as np
import unittest


import teneva


import tt
from tt.cross.rectcross import cross as ttpy_cross


np.random.seed(42)


def _err(A, B):
    if isinstance(A, list):
        A = tt.tensor.from_list(A)
    if isinstance(B, list):
        B = tt.tensor.from_list(B)
    if isinstance(A, (int, float)):
        return abs((A - B) / A)
    if isinstance(A, np.ndarray):
        return np.linalg.norm(A - B) / np.linalg.norm(A)
    return (A - B).norm() / A.norm()


class TestCross(unittest.TestCase):

    def test_base(self):
        N = [9, 8, 8, 9, 8]  # Shape of the tensor
        d = len(N)           # Dimension of the problem
        M = 10000            # Number of test cases
        nswp     = 6         # Sweep number
        eps      = 1.E-6     # Desired accuracy
        kickrank = 2         # Cross parameter
        rf       = 2         # Cross parameter

        def f(x):
            y = np.sum([x[i]**(i+1) for i in range(d)])
            y = np.sin(y)**2 / (1 + y**2)
            return  y

        def f_vect(X):
            """Naively vectorized model function."""
            return np.array([f(x) for x in X])

        X_tst = np.vstack([np.random.choice(N[i], M) for i in range(d)]).T
        Y_tst = np.array([f(x) for x in X_tst])

        Y0 = teneva.rand(N, 2)

        Y = teneva.cross(f_vect, Y0, nswp, kickrank, rf)
        Y = teneva.truncate(Y, eps)

        get = teneva.getter(Y)
        Z = np.array([get(x) for x in X_tst])
        e1 = np.linalg.norm(Z - Y_tst) / np.linalg.norm(Y_tst)
        r1 = teneva.erank(Y)

        def solver_cro(f, Y0, nswp=10, eps=None, kickrank=2, rf=2):
            Y0 = tt.tensor.from_list(Y0)
            Y = ttpy_cross(myfun=f, x0=Y0, nswp=nswp, eps=1.E-16,
                      eps_abs=0., kickrank=kickrank, rf=rf,
                      verbose=False, stop_fun=None, approx_fun=None)
            if eps:
                Y = Y.round(eps)
            return tt.tensor.to_list(Y)

        Yr_cro = solver_cro(f_vect, Y0, nswp, eps, kickrank, rf)

        get = teneva.getter(Yr_cro)
        Z = np.array([get(x) for x in X_tst])
        e2 = np.linalg.norm(Z - Y_tst) / np.linalg.norm(Y_tst)
        r2 = teneva.erank(Y)

        self.assertTrue(_err(e1, e2) < 5.E-6)
        self.assertTrue(_err(r1, r2) < 5.E-14)


class TestErank(unittest.TestCase):

    def setUp(self):
        self.Z = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Y = tt.tensor.to_list(self.Z)
        self.e = 1.E-14

    def test_base(self):
        m1 = self.Z.erank
        m2 = teneva.erank(self.Y)

        self.assertTrue(_err(m1, m2) < self.e)


class TestGetCdf(unittest.TestCase):

    def test_base(self):

        def cdf0(x, X):
            return np.array([np.sum(X <= p) for p in x]) / len(X)

        X = [3, 3, 1, 4]
        x = np.array([3, 55, 0.5, 1.5, 2])

        ecdf = teneva.get_cdf(X)

        v1 = ecdf(x)
        v2 = cdf0(x, X)

        self.assertTrue(_err(v1, v2) < 1.E-16)


class TestGet(unittest.TestCase):

    def setUp(self):
        self.Z = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Y = tt.tensor.to_list(self.Z)
        self.x = np.array([0, 1, 2, 1, 0])
        self.e = 1.E-14

    def test_base(self):
        y1 = self.Z[self.x]
        y2 = teneva.get(self.Y, self.x)

        self.assertTrue(_err(y1, y2) < self.e)

    def test_getter(self):
        get = teneva.getter(self.Y)

        y1 = self.Z[self.x]
        y2 = get(self.x)

        self.assertTrue(_err(y1, y2) < self.e)


class TestMean(unittest.TestCase):

    def setUp(self):
        self.Z = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Y = tt.tensor.to_list(self.Z)
        self.e = 1.E-14

    def test_base(self):
        m1 = tt.sum(self.Z) / np.prod([G.shape[1] for G in self.Y])
        m2 = teneva.mean(self.Y)

        self.assertTrue(_err(m1, m2) < self.e)


class TestMul(unittest.TestCase):

    def setUp(self):
        self.Z1 = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Z2 = tt.rand(np.array([5, 6, 7, 8, 9]), r=12)
        self.Y1 = tt.tensor.to_list(self.Z1)
        self.Y2 = tt.tensor.to_list(self.Z2)
        self.e = 1.E-14

    def test_base(self):
        C1 = self.Z1 * self.Z2
        C2 = teneva.mul(self.Y1, self.Y2)

        self.assertTrue(_err(C1, C2) < self.e)


class TestNorm(unittest.TestCase):

    def setUp(self):
        self.Z = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Y = tt.tensor.to_list(self.Z)
        self.e = 1.E-14

    def test_base(self):
        n1 = self.Z.norm()
        n2 = teneva.norm(self.Y)

        self.assertTrue(_err(n1, n2) < self.e)


class TestRand(unittest.TestCase):

    def setUp(self):
        self.e = 1.E-15

    def test_base(self):
        def sf(k):
            return np.array([0.1] * k)

        n = np.array([5, 6, 7, 8, 9])
        r = 4
        Z = tt.rand(n, r=r, samplefunc=sf)
        Y = teneva.rand(n, r, sf)

        self.assertTrue(_err(Z, Y) < self.e)


class TestRecap(unittest.TestCase):

    def setUp(self):
        self.Z = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Y = tt.tensor.to_list(self.Z)
        self.e = 1.E-14

    def test_base(self):
        m1 = tt.sum(self.Z)
        m2 = teneva.recap(self.Y)

        self.assertTrue(_err(m1, m2) < self.e)


class TestTruncate(unittest.TestCase):

    def setUp(self):
        self.Z = tt.rand(np.array([5, 6, 7, 8, 9]), r=10)
        self.Y = tt.tensor.to_list(self.Z)
        self.e = 1.E-14

    def test_base(self):
        e = 1.E-2
        Z = self.Z.round(e)
        Y = teneva.truncate(self.Y, e)

        self.assertTrue(_err(Z, Y) < self.e)


if __name__ == '__main__':
    unittest.main()



import numba as nb
import numpy as np


from .utils import orthogonalize
from .utils import svd_truncated
from .utils import unfolding_right


def erank(Y):
    d = len(Y)
    N = np.array([G.shape[1] for G in Y])
    R = np.array([1] + [G.shape[-1] for G in Y[:-1]] + [1])

    sz = np.dot(N * R[0:d], R[1:])
    b = R[0] * N[0] + N[d - 1] * R[d]
    if d is 2:
        er = sz * 1. / b
    else:
        a = np.sum(N[1:d - 1])
        er = (np.sqrt(b * b + 4 * a * sz) - b) / (2 * a)
    return er


def get(Y, x):
    Q = Y[0][0, x[0], :]
    for i in range(1, len(Y)):
        Q = np.einsum('q,qp->p', Q, Y[i][:, x[i], :])
    return Q[0]


def getter(Y, compile=True):
    Y_nb = tuple([np.array(G, order='F') for G in Y])

    @nb.jit(nopython=True)
    def get(x):
        Q = Y_nb[0]
        y = [Q[0, x[0], r2] for r2 in range(Q.shape[2])]
        for i in range(1, len(Y_nb)):
            Q = Y_nb[i]
            R = np.zeros(Q.shape[2])
            for r1 in range(Q.shape[0]):
                for r2 in range(Q.shape[2]):
                    R[r2]+= y[r1] * Q[r1, x[i], r2]
            y = list(R)
        return y[0]

    if compile:
        y = get(np.zeros(len(Y_nb), dtype=int)) # Compile

    return get


def mean(Y, P=None, norm=True):
    R = np.ones((1, 1))
    for i in range(len(Y)):
        n = Y[i].shape[1]
        if P is not None:
            Q = P[i, 0:n]
        else:
            Q = np.ones(n) / n if norm else np.ones(n)
        R = R @ np.einsum('rmq,m->rq', Y[i], Q)
    return R[0, 0]


def mul(A, B):
    C = []
    for G1, G2 in zip(A, B):
        G = G1[:, None, :, :, None] * G2[None, :, :, None, :]
        G = G.reshape([G1.shape[0]*G2.shape[0], -1, G1.shape[-1]*G2.shape[-1]])
        C.append(G)
    return C


def norm(Y):
    return np.sqrt(recap(mul(Y, Y)))


def rand(N, R, f=np.random.randn):
    N = np.asanyarray(N, dtype=np.int32)
    d = N.size
    if d < 3:
        raise ValueError('Dimension should be at least 3.')

    if isinstance(R, (int, float)):
        R = [1] + [int(R)] * (d - 1) + [1]
    R = np.asanyarray(R, dtype=np.int32)

    ps = np.cumsum(np.concatenate(([1], N * R[0:d] * R[1:d +1])))
    ps = ps.astype(np.int32)
    core = f(ps[d] - 1)

    Y = []
    for i in range(d):
        G = core[ps[i]-1:ps[i+1]-1]
        Y.append(G.reshape((R[i], N[i], R[i+1]), order='F'))
    return Y


def recap(Y):
    return mean(Y, norm=False)


def truncate(Y, e, rmax=np.iinfo(np.int32).max):
    d = len(Y)
    N = [G.shape[1] for G in Y]
    orthogonalize(Y, d-1)
    delta = e / np.sqrt(d-1) * np.linalg.norm(Y[-1])
    for mu in range(d-1, 0, -1):
        M = unfolding_right(Y[mu])
        L, M = svd_truncated(M, delta=delta, rmax=rmax, left_ortho=False)
        Y[mu] = np.reshape(M, [-1, N[mu], Y[mu].shape[2]], order='F')
        Y[mu-1] = np.einsum('ijk,kl', Y[mu-1], L, optimize=True)
    return Y



import numpy as np
import unittest


import teneva


np.random.seed(42)


class TestCross(unittest.TestCase):

    def test_base(self):
        N = [9, 8, 8, 9, 8]  # Shape of the tensor
        d = len(N)           # Dimension of the problem
        M = 10000            # Number of test cases
        nswp     = 10        # Sweep number
        eps      = 1.E-6     # Desired accuracy
        kickrank = 1         # Cross parameter
        rf       = 1         # Cross parameter

        def f(x):
            y = np.sum([x[i]**(i+1) for i in range(d)])
            y = np.sin(y)**2 / (1 + y**2)
            return  y

        def f_vect(X):
            """Naively vectorized model function."""
            return np.array([f(x) for x in X])

        X_tst = np.vstack([np.random.choice(N[i], M) for i in range(d)]).T
        Y_tst = np.array([f(x) for x in X_tst])

        Y0 = teneva.rand(N, 2)

        Y = teneva.cross(f_vect, Y0, nswp, kickrank, rf)
        Y = teneva.truncate(Y, eps)

        get = teneva.getter(Y)
        Z = np.array([get(x) for x in X_tst])
        e = np.linalg.norm(Z - Y_tst) / np.linalg.norm(Y_tst)
        r = teneva.erank(Y)

        self.assertTrue(e < 5.E-7)
        self.assertTrue(r < 13.81)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from scipy.linalg import lu
from scipy.linalg import solve_triangular


def maxvol(A, e=1.05, K=100):
    N, r = A.shape
    assert e >= 1 and N > r and K > 0
    P, L, U = lu(A)
    I = P.argmax(axis=0)
    Q = solve_triangular(U, A.T, trans=1, lower=False)
    B = solve_triangular(L[:r, :], Q, trans=1, unit_diagonal=True, lower=True)
    for _ in range(K):
        i, j = np.divmod(np.abs(B).argmax(), N)
        if np.abs(B[i, j]) <= e: break
        x = B[i, :]
        y = B[:, j].copy()
        y[i]-= 1.
        I[i] = j
        B-= np.outer(y / B[i, j], x)
    return I[:r], B.T


def rect_maxvol(A, e, N_min, N_max, e0=1.05, K0=10):
    N, r = A.shape
    assert e >= 1 and N > r and N_min >= r and N_min <= N_max and N_max <= N
    I_tmp, B = maxvol(A, e0, K0)
    I = np.hstack([I_tmp, np.zeros(N_max-r, dtype=I_tmp.dtype)])
    S = np.ones(N, dtype=np.int32)
    S[I_tmp] = 0
    F = S * np.linalg.norm(B, axis=1)**2
    for k in range(r, N_max):
        i = np.argmax(F)
        if k >= N_min and F[i] <= e*e: break
        I[k] = i
        S[i] = 0
        v = B.dot(B[i])
        l = 1. / (1 + v[i])
        B = np.hstack([B - l * np.outer(v, B[i]), l * v.reshape(-1, 1)])
        F = S * (F - l * v * v)
    I = I[:B.shape[1]]
    B[I] = np.eye(B.shape[1], dtype=B.dtype)
    return I, B
